'''
Helper module
'''
import json
import logging
import sys

from termcolor import colored
from six import StringIO, print_


class Logger:
    ch = logging.NullHandler()
    quiet = False

    @staticmethod
    def set(ch, quiet):
        Logger.ch = ch
        Logger.quiet = quiet
        if Logger.quiet:
            sys.stdout = StringIO()

    @staticmethod
    def get(name_class, level=logging.DEBUG):
        # print_('get logger - class:', name_class, Logger.ch)
        logger = logging.getLogger(name_class)
        logger.setLevel(level)
        logger.addHandler(Logger.ch)
        assert isinstance(logger, logging.Logger)
        return logger

    @staticmethod
    def print_(*args):
        if not Logger.quiet:
            print_(*args, end='', flush=True)

    @staticmethod
    def print_error(*args):
        if not Logger.quiet:
            print_(colored('ERROR: ' + ' '.join(args), 'red'))

    @staticmethod
    def println(*args):
        if not Logger.quiet:
            print_(*args)


def get_consol_handler():
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter((
        '%(levelname) -3s %(asctime)s %(name)'
        '-3s %(funcName)'
        '-1s %(lineno) -0s: %(message)s'
    ))
    ch.setFormatter(formatter)
    return ch


def get_attributes(args, nodes):
    get = nodes
    for a in args:
        get = get[a]
    return get


def print_TOSCA(tosca, indent=2):
    space = ' ' * indent

    def _rec_print(item, tab, res):
        if type(item) is dict:
            for key, value in item.items():
                if type(value) is str or \
                   (type(value) is dict and 'get_input' in value):
                    return res + tab + str(key) + ': ' + str(value)
                else:
                    res += tab + str(key) + ':\n'
                    return _rec_print(value, tab + space, res)
        elif type(item) is list:
            for value in item:
                if type(value) is str:
                    return res + tab + '- ' + value + '\n'
                else:
                    key, value = list(value.items())[0]
                    res += res + tab + '- ' + key + ':\n'
                    return _rec_print(value, tab + space + '  ', res)
    res = ''
    if hasattr(tosca, 'nodetemplates'):
        if tosca.inputs:
            res += "\ninputs:\n"
            for input in tosca.inputs:
                res += space + input.name

        nodetemplates = tosca.nodetemplates
        if nodetemplates:
            res += "\nnodetemplates:\n"
            for node in nodetemplates:
                res += space + node.name + '\n'
                res += _rec_print(node.entity_tpl, space + space, '')
                res += '\n'
    return res


def print_json(stream, fprint):
    for line in stream:
        fprint('\t' + json.dumps(json.loads(line.decode("utf-8")), indent=2))


def print_byte(stream, fprint):
    for line in stream:
        fprint('\t' + line.decode("utf-8").strip())



'''
Docker interface module
'''
import os
from functools import wraps
from os import path

import six
from docker import APIClient, errors

from . import helper
from .graph.artifacts import Dockerfile
from .graph.nodes import Container, Root, Volume
from .helper import Logger

_cli = None


def _get_name(func):
    @wraps(func)
    def func_wrapper(*args, **kwds):
        if isinstance(args[0], six.string_types):
            return func(*args, **kwds)
        else:
            assert isinstance(args[0], (Container, Volume))
            return func(args[0].full_name, *args[1:], **kwds)
    return func_wrapper


def _inject_docker_cli(func):
    @wraps(func)
    def func_wrapper(*args, **kwds):
        global _cli
        if _cli is None:
            _cli = APIClient(base_url=os.environ.get('DOCKER_HOST'))
        return func(_cli, *args, **kwds)
    return func_wrapper


def _get_tmp_dir(node):
    assert isinstance(node, Root)
    return path.join(node.tpl.tmp_dir, node.name)


@_inject_docker_cli
def create_container(_cli,
                     con,
                     cmd=None,
                     entrypoint=None,
                     from_saved=False,
                     force=False):
    _log = Logger.get(__name__)

    def create():
        tmp_dir = _get_tmp_dir(con)
        try:
            os.makedirs(tmp_dir)
        except Exception:
            pass
        img_name = con.image.format
        _log.debug('image_name: {}'.format(img_name))

        if from_saved:
            saved_image = get_saved_image(con)
            if inspect_image(saved_image):
                img_name = saved_image

        # _log.debug('container: {}'.format(con.get_str_obj()))

        con.id = _cli.create_container(
            name=con.full_name,
            image=img_name,
            user='root',
            entrypoint=entrypoint,
            command=cmd if cmd else con.cmd,
            environment=con.env,
            detach=True,
            # stdin_open=True,
            ports=[key for key in con.ports.keys()]
            if con.ports else None,
            volumes=['/tmp/dt'] + [k for k, v in con.volume] +\
                    [k for k in con.share_data.keys()],
            networking_config=_cli.create_networking_config({
                _get_app_name(con): _cli.create_endpoint_config(
                    links=con.connection
                )}),
            host_config=_cli.create_host_config(
                port_bindings=con.ports,
                binds=[tmp_dir + ':/tmp/dt'] +\
                      [v + ':' + k for k, v in con.volume] +\
                      ['{}:{}'.format(v, k) for k, v in con.share_data.items()],
            )
        ).get('Id')

    assert isinstance(con, Container)

    if isinstance(con.image, Dockerfile):
        _log.debug('start building..')
        build_image(con)
        _log.debug('stop building..')
    elif not from_saved:
        _log.debug('start pulling.. {}'.format(con.image))
        # helper.print_json(
        _cli.pull(con.image.format)
        # , _log.debug)
        _log.debug('end pulling..')

    try:
        create()
    except errors.APIError as e:
        if force:
            delete_container(con)
            create()
        else:
            _log.debug(e)
            raise e


# @_inject_docker_cli
# def pull_image(_cli, image):
#     assert isinstance(image, six.string_types)
#     _cli.pull(image)

@_get_name
@_inject_docker_cli
def stop_container(_cli, name):
    _log = Logger.get(__name__)
    try:
        _cli.stop(name)
    except errors.NotFound as e:
        _log.error(e)
        raise e


@_get_name
@_inject_docker_cli
def start_container(_cli, name, wait=False):
    _log = Logger.get(__name__)
    try:
        _cli.start(name)
        if wait:
            _log.debug('wait container..')
            _cli.wait(name)
            # helper.print_byte(
            #     _cli.logs(name, stream=True),
            #     _log.debug
            # )
            _log.debug('end wait container..')
    except errors.NotFound as e:
        _log.error(e)
        raise e


@_get_name
@_inject_docker_cli
def delete_container(_cli, name):
    _log = Logger.get(__name__)
    try:
        _cli.remove_container(name, v=True)
    except (errors.NotFound, errors.APIError) as e:
        _log.error(e)
        raise e


@_get_name
@_inject_docker_cli
def exec_cmd(_cli, name, cmd):
    _log = Logger.get(__name__)
    if not is_running(name):
        raise Exception('{} is not running'.format(name))
    try:
        exec_id = _cli.exec_create(name, cmd,
                                   stdout=False,
                                   stderr=False)
        status = _cli.exec_start(exec_id)
        _log.debug(status)

        check = 'rpc error:' != status[:10].decode("utf-8")
        _log.debug('check: {}'.format(check))
        if not check:
            raise errors.APIError
    except errors.APIError as e:
        _log.error(e)
        raise e


@_inject_docker_cli
def build_image(_cli, node):
    assert isinstance(node, Container)
    # helper.print_json(
    return _cli.build(
        path='/'.join(node.image.dockerfile.split('/')[0:-1]),
        dockerfile='./' + node.image.dockerfile.split('/')[-1],
        tag=node.image.name,
        pull=True,
        quiet=True
    )
    # )


@_inject_docker_cli
def create_volume(_cli, volume):
    assert isinstance(volume, Volume)
    _log = Logger.get(__name__)
    _log.debug('volume opt: {}'.format(volume.get_all_opt()))
    return _cli.create_volume(
        volume.full_name, 'local', volume.get_all_opt()
    )


@_get_name
@_inject_docker_cli
def delete_volume(_cli, name):
    return _cli.remove_volume(name)


@_inject_docker_cli
def get_containers(_cli, all=False):
    return _cli.containers(all=all)


@_inject_docker_cli
def get_volumes(_cli):
    volumes = _cli.volumes()
    return volumes['Volumes'] or []


def inspect(item):
    return (inspect_image(item) or
            inspect_container(item) or
            inspect_volume(item))


@_inject_docker_cli
def inspect_image(_cli, name):
    assert isinstance(name, six.string_types)
    try:
        return _cli.inspect_image(name)
    except errors.NotFound:
        return None


@_get_name
@_inject_docker_cli
def inspect_container(_cli, name):
    try:
        return _cli.inspect_container(name)
    except errors.NotFound:
        return None


@_get_name
@_inject_docker_cli
def inspect_volume(_cli, name):
    try:
        return _cli.inspect_volume(name)
    except errors.NotFound:
        return None


def remove_all_containers():
    for c in get_containers(all=True):
        stop_container(c['Id'])
        delete_container(c['Id'])


def remove_all_volumes():
    for v in get_volumes():
        delete_volume(v['Name'])


@_inject_docker_cli
def create_network(_cli, name, subnet='172.25.0.0/16'):
    _log = Logger.get(__name__)
    # docker network create -d bridge --subnet 172.25.0.0/16 isolated_nw
    # self.delete_network(name)
    try:
        _cli.create_network(name=_get_app_name(name),
                            driver='bridge',
                            ipam={'subnet': subnet},
                            check_duplicate=True)
    except errors.APIError:
        _log.debug('network already exists!')


@_inject_docker_cli
def delete_network(_cli, name):
    _log = Logger.get(__name__)
    try:
        _cli.remove_network(_get_app_name(name))
    except errors.APIError:
        _log.debug('network not exists!')


@_inject_docker_cli
def delete_image(_cli, name):
    assert isinstance(name, six.string_types)
    try:
        _cli.remove_image(name)
    except errors.NotFound:
        pass


@_inject_docker_cli
def update_container(_cli, node, cmd):
    assert isinstance(node, Container)
    stat = inspect_image(node.image.format)
    old_cmd = stat['Config']['Cmd'] or None
    old_entry = stat['Config']['Entrypoint'] or None

    # if is_running(node):
    #     stop_container(node)
    #     delete_container(node)

    create_container(node,
                     cmd=cmd,
                     entrypoint='',
                     from_saved=True,
                     force=True)

    start_container(node.id, wait=True)
    # stop_container(node.id)

    _cli.commit(node.id, get_saved_image(node))

    # stop_container(node)
    # delete_container(node)
    create_container(node,
                     cmd=node.cmd or old_cmd,
                     entrypoint=old_entry,
                     from_saved=True,
                     force=True)

    _cli.commit(node.id, get_saved_image(node))


def is_running(container):
    _log = Logger.get(__name__)
    stat = inspect_container(container)
    stat = stat is not None and stat['State']['Running'] is True
    _log.debug('State: {}'.format(stat))
    return stat


def get_saved_image(node):
    assert isinstance(node, Container)
    return '{}/{}'.format(_get_app_name(node), node.name.lower())


def _get_app_name(node):
    assert isinstance(node, (six.string_types, Container))
    if isinstance(node, Container):
        node = node.tpl.name
    return 'tosker_{}'.format(node.lower())


# @property
# def tmp_dir(self):
#     return self._tmp_dir



import unittest

from tosker import docker_interface as docker
# from tosker import helper
from tosker.orchestrator import Orchestrator
from tosker.storage import Memory
from tosker.tosca_parser import get_tosca_template


class Test_Orchestrator(unittest.TestCase):
    @staticmethod
    def setUpClass():
        pass

    def setUp(self):
        docker.remove_all_containers()
        docker.remove_all_volumes()

        self.orchestrator = Orchestrator(
            # log_handler=helper.get_consol_handler()
        )
        Memory.purge()

    def get_tpl(self):
        if not hasattr(self, 'tpl'):
            self.tpl = get_tosca_template(self.file)
        return self.tpl

    def create(self):
        self.orchestrator.orchestrate(self.file, ['create'])
        for c in self.get_tpl().containers:
            self.assertIsNotNone(
                docker.inspect_container(c)
            )

        for c in self.get_tpl().volumes:
            self.assertIsNotNone(
                docker.inspect_volume(c)
            )

    def start(self):
        self._start(check=lambda x: x['State']['Running'])

    def start_check_exit(self):
        self._start(check=lambda x: x['State']['ExitCode'] == 0)

    def _start(self, check):
        self.orchestrator.orchestrate(self.file, ['start'])
        for c in self.get_tpl().containers:
            stat = docker.inspect_container(c)
            self.assertIsNotNone(stat)
            self.assertTrue(check(stat))

    def stop(self):
        self.orchestrator.orchestrate(self.file, ['stop'])
        for c in self.get_tpl().containers:
            stat = docker.inspect_container(c)
            # print('DEBUG: ', stat)
            self.assertIsNotNone(stat)
            self.assertFalse(stat['State']['Running'])

        for c in self.get_tpl().volumes:
            self.assertIsNotNone(
                docker.inspect_volume(c)
            )

    def delete(self):
        self.orchestrator.orchestrate(self.file, ['delete'])
        for c in self.get_tpl().containers:
            self.assertIsNone(
                docker.inspect_container(c)
            )

        for c in self.get_tpl().volumes:
            self.assertIsNotNone(
                docker.inspect_volume(c)
            )



import unittest

from tosker import ui


class Test_Ui(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        pass

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_args_flags(self):
        args = ['--help', '--debug', '--quiet', '--version']
        error, mod,\
            file, cmds, comps, flags, inputs = ui._parse_unix_input(args)
        self.assertIsNone(error)
        self.assertTrue(flags.get('help', False))
        self.assertTrue(flags.get('debug', False))
        self.assertTrue(flags.get('quiet', False))
        self.assertTrue(flags.get('version', False))

        args = ['-v', '-h', '-q']
        error, mod,\
            file, cmds, comps, flags, inputs = ui._parse_unix_input(args)
        self.assertIsNone(error)
        self.assertTrue(flags.get('quiet', False))
        self.assertTrue(flags.get('help', False))
        self.assertTrue(flags.get('version', False))

        args = ['--test']
        error, mod,\
            file, cmds, comps, flags, inputs = ui._parse_unix_input(args)
        self.assertIsNotNone(error)

        args = ['-t']
        error, mod,\
            file, cmds, comps, flags, inputs = ui._parse_unix_input(args)
        self.assertIsNotNone(error)

    def test_args_command(self):
        args = ['data/examples/hello.yaml',
                'create', 'start', 'stop', 'delete']
        error, mod,\
            file, cmds, comps, flags, inputs = ui._parse_unix_input(args)
        self.assertIsNone(error)
        self.assertEqual('data/examples/hello.yaml', file)
        self.assertListEqual(['create', 'start', 'stop', 'delete'], cmds)

        args = ['data/examples/hello.yaml',
                'comp1', 'comp2', 'comp3',
                'create', 'start', 'stop', 'delete']
        error, mod,\
            file, cmds, comps, flags, inputs = ui._parse_unix_input(args)
        self.assertIsNone(error)
        self.assertEqual('data/examples/hello.yaml', file)
        self.assertListEqual(['create', 'start', 'stop', 'delete'], cmds)
        self.assertListEqual(['comp1', 'comp2', 'comp3'], comps)

        args = ['data/examples/hello.yaml',
                'create', 'test', 'start', 'stop', 'delete']
        error, mod,\
            file, cmds, comps, flags, inputs = ui._parse_unix_input(args)
        self.assertIsNotNone(error)

        args = ['asd',
                'comp1', 'comp2', 'comp3',
                'create', 'start', 'stop', 'delete']
        error, mod,\
            file, cmds, comps, flags, inputs = ui._parse_unix_input(args)
        self.assertIsNotNone(error)

    def test_args_inputs(self):
        args = ['--input1', 'value1', '--input2', 'value2']
        error, mod,\
            file, cmds, comps, flags, inputs = ui._parse_unix_input(args)
        self.assertIsNone(error)
        self.assertEqual(inputs.get('input1'), 'value1')
        self.assertEqual(inputs.get('input2'), 'value2')

        args = ['--input1', '--input2', 'value2']
        error, mod,\
            file, cmds, comps, flags, inputs = ui._parse_unix_input(args)
        self.assertIsNotNone(error)

    def test_check_file(self):
        self.assertIsNotNone(
            ui._check_file('data/examples/hello.yaml')
        )
        self.assertIsNotNone(
            ui._check_file(
                'data/examples/node-mongo-csar/node-mongo.csar')
        )
        self.assertIsNone(
            ui._check_file(
                'data/examples/node-mongo-csar')
        )
        self.assertIsNone(
            ui._check_file(
                'data/examples/node-mongo-csar/'
                'node-mongo/scripts/install_node.sh')
        )



'''
Container manager module
'''
# from ..helper import Logger
# from ..graph.artifacts import Dockerfile
from .. import docker_interface
from ..graph.nodes import Container


class Container_manager:

    @staticmethod
    def create(node):
        assert isinstance(node, Container)
        # if node.executable:
        docker_interface.create_container(node)
        # else:
        #     if isinstance(node.image, Dockerfile):
        #         docker_interface.build_image(node)
        #     else:
        #         docker_interface.pull_image(node.image.format)

    @staticmethod
    def start(node):
        assert isinstance(node, Container)
        if not node.executable:
            return
        stat = docker_interface.inspect_container(node)
        if stat is not None:
            node.id = stat['Id']
            docker_interface.start_container(node)

    @staticmethod
    def stop(node):
        assert isinstance(node, Container)
        docker_interface.stop_container(node)
        # docker_interface.delete_container(node)
        docker_interface.create_container(node, from_saved=True, force=True)

    @staticmethod
    def delete(node):
        assert isinstance(node, Container)
        docker_interface.delete_container(node)
        docker_interface.delete_image(docker_interface.get_saved_image(node))



'''
Relationships module
'''
import six


class Relationship(object):

    def __init__(self, origin, to):
        self.origin = origin
        self.to = to

    def __str__(self):
        return 'Relationship'


def _get_str_name(obj):
    return obj if isinstance(obj, six.string_types) else obj.name


def _get_str_full_name(obj):
    return obj if isinstance(obj, six.string_types) else obj.full_name


class ConnectsTo(Relationship):

    def __init__(self, origin, node, alias=None):
        super(self.__class__, self).__init__(origin, node)
        self.alias = alias

    @property
    def format(self):
        full_name = _get_str_full_name(self.to)
        if self.alias is not None:
            return (full_name, self.alias)
        else:
            return (full_name, _get_str_name(self.to))

    def __str__(self):
        return 'ConnectsTo'


class HostedOn(Relationship):

    def __init__(self, origin, node):
        super(self.__class__, self).__init__(origin, node)

    @property
    def format(self):
        return _get_str_name(self.to)

    def __str__(self):
        return 'HostedOn'


class AttachesTo(Relationship):

    def __init__(self, origin, node, folder=None):
        super(self.__class__, self).__init__(origin, node)
        self.location = folder

    @property
    def format(self):
        if self.location is not None:
            return (self.location, _get_str_full_name(self.to))
        else:
            return _get_str_full_name(self.to)

    def __str__(self):
        return 'AttachesTo'


class DependsOn(Relationship):

    def __init__(self, origin, node):
        super(self.__class__, self).__init__(origin, node)

    @property
    def format(self):
        return _get_str_name(self.to)

    def __str__(self):
        return 'DependsOn'



'''
Template module
'''
import six

from .nodes import Container, Root, Software, Volume


class Template:

    def __init__(self, name):
        self._nodes = {}
        self.name = name
        self.outputs = []
        self.tmp_dir = None

    @property
    def nodes(self):
        return (v for k, v in self._nodes.items())

    @property
    def containers(self):
        return (v for k, v in self._nodes.items() if type(v) is Container)

    @property
    def volumes(self):
        return (v for k, v in self._nodes.items() if type(v) is Volume)

    @property
    def software(self):
        return (v for k, v in self._nodes.items() if type(v) is Software)

    def push(self, node):
        self._nodes[node.name] = node

    def __getitem__(self, name):
        return self._nodes.get(name, None)

    def __contains__(self, b):
        if isinstance(b, six.string_types):
            return self[b] is not None
        elif isinstance(b, Root):
            return self[b.name] is not None
        else:
            return False

    def __str__(self):
        return ', '.join((i.name for i in self.nodes))



'''
Orchestrator module
'''
import logging
import os
import shutil
import traceback
from functools import wraps
from glob import glob

import six
from halo import Halo
from tabulate import tabulate
from termcolor import colored
from yaml.scanner import ScannerError
from toscaparser.common.exception import ValidationError

from . import docker_interface, helper
from .graph.nodes import Container, Software, Volume
from .helper import Logger
from .managers.container_manager import Container_manager
from .managers.software_manager import Software_manager
from .managers.volume_manager import Volume_manager
from .storage import Memory
from .tosca_parser import get_tosca_template

try:
    from os import scandir
except ImportError:
    from scandir import scandir


def _filter_components(*comps):
    def _filter_components_decorator(func):
        @wraps(func)
        def func_wrapper(self, *args):
            filter_componets = [c for c in args[0]
                                if isinstance(c, comps)]
            return func(self, filter_componets, *args[1:])
        return func_wrapper
    return _filter_components_decorator


def _filter_interface(interface):
    def _filter_interface_decorator(func):
        @wraps(func)
        def func_wrapper(self, *args):
            filter_componets = [c for c in args[0]
                                if interface in c.interfaces]
            return func(self, filter_componets, *args[1:])
        return func_wrapper
    return _filter_interface_decorator


class Orchestrator:

    def __init__(self,
                 log_handler=logging.NullHandler(),
                 quiet=True,
                 tmp_dir='/tmp/tosker',
                 data_dir='/tmp/tosker'):  # TODO: use /usr/lib/tokser instead
        Logger.set(log_handler, quiet)
        self._log = Logger.get(__name__)
        self._tmp_dir = tmp_dir

        # Setup Storage system (folder and class
        self._data_dir = data_dir
        try:
            os.makedirs(data_dir)
        except os.error as e:
            pass
        Memory.set_db(data_dir)

        status, faulty = self._update_state()
        Logger.println('update memory {} {}'.format(
            'ok' if status else 'fixed',
            '({})'.format(', '.join(faulty)) if not status else ''))

    def orchestrate(self, file_path, commands, components=[], inputs={}):
        # Parse TOSCA file
        try:
            tpl = get_tosca_template(file_path, inputs)
        except ScannerError as e:
            Logger.print_error('YAML parse error\n    {}'.format(e))
            return False
        except ValidationError as e:
            Logger.print_error('TOSCA validation error\n    {}'.format(e))
            return False
        except ValueError as e:
            Logger.print_error('TosKer validation error\n    {}'.format(e))
            return False
        except Exception as e:
            Logger.print_error('Internal error\n    {}'.format(e))
            self._log.debug('Exception type: {}'.format(type(e)))
            self._log.debug(colored(traceback.format_exc(), 'red'))
            return False

        # Check if inputs components exists in the TOSCA file
        if not self._components_exists(tpl, components):
            Logger.print_error('a selected component do not exists')
            return False

        # Create temporany directory
        tpl.tmp_dir = os.path.join(self._tmp_dir, tpl.name)
        try:
            os.makedirs(tpl.tmp_dir)
        except os.error as e:
            self._log.info(e)

        # Start orchestration
        try:
            if len(components) == 0:
                components = [n.name for n in tpl.nodes]

            # must calculate the down extension/sorting
            if any((c in ('create', 'start') for c in commands)):
                down_extension = self._extend_down(tpl, components)
                down_deploy = self._sort(tpl, down_extension)
                self._log.debug('down_deploy: {}'.format(', '
                                ''.join((c.name for c in down_deploy))))
            # must calculate the up extension/sorting
            if any((c in ('stop', 'delete') for c in commands)):
                up_extension = self._extend_up(tpl, components)
                up_deploy = list(reversed(self._sort(tpl, up_extension)))
                self._log.debug('up_deploy: {}'.format(', '
                                ''.join((c.name for c in up_deploy))))

            for cmd in commands:
                {
                  'create': lambda: self._create(down_deploy, tpl),
                  'start': lambda: self._start(down_deploy, tpl),
                  'stop': lambda: self._stop(up_deploy, tpl),
                  'delete': lambda: self._delete(up_deploy, tpl),
                }.get(cmd)()

            self._print_outputs(tpl)
        except Exception as e:
            self._log.debug('Exception type: {}'.format(type(e)))
            self._log.debug(traceback.format_exc())
            self._print_cross(e)
            return False
        return True

    # @_filter_interface('create')
    def _create(self, components, tpl):
        self._print_loading_start('Create network... ')
        docker_interface.create_network(tpl.name)  # TODO: da rimuovere
        self._print_tick()
        for node in components:
            self._print_loading_start('Create {}... '.format(node))

            status = Memory.get_comp_state(node)
            if Memory.STATE.DELETED == status:
                if isinstance(node, Container):
                    Container_manager.create(node)
                elif isinstance(node, Volume):
                    Volume_manager.create(node)
                elif isinstance(node, Software):
                    Software_manager.create(node)
                    Software_manager.configure(node)

                Memory.update_state(node, Memory.STATE.CREATED)

                self._print_tick()
            else:
                self._print_skip()
                self._log.info('skipped already created')

    @_filter_interface('start')
    def _start(self, components, tpl):
        for node in components:
            self._print_loading_start('Start {}... '.format(node))

            status = Memory.get_comp_state(node)
            if Memory.STATE.STARTED == status:
                self._print_skip()
                self._log.info('skipped already started')
            elif Memory.STATE.CREATED == status or\
                    'create' not in node.interfaces:
                if isinstance(node, Container):
                    Container_manager.start(node)
                elif isinstance(node, Software):
                    Software_manager.start(node)
                Memory.update_state(node, Memory.STATE.STARTED)
                self._print_tick()
            else:
                self._print_cross('the components must be created first')
                self._log.info('{} have to be created first'.format(node))
                break

    @_filter_interface('stop')
    def _stop(self, components, tpl):
        for node in components:
            self._print_loading_start('Stop {}... '.format(node))

            status = Memory.get_comp_state(node)
            if Memory.STATE.STARTED == status:
                if isinstance(node, Container):
                    Container_manager.stop(node)
                elif isinstance(node, Software):
                    Software_manager.stop(node)
                Memory.update_state(node, Memory.STATE.CREATED)
                self._print_tick()
            else:
                self._print_skip()
                self._log.info('skipped already stopped')

    @_filter_interface('delete')
    def _delete(self, components, tpl):
        self._log.debug('start delete')
        for node in components:
            self._print_loading_start('Delete {}... '.format(node))

            status = Memory.get_comp_state(node)
            if Memory.STATE.CREATED == status:
                if isinstance(node, Container):
                    Container_manager.delete(node)
                elif isinstance(node, Software):
                    Software_manager.delete(node)
                Memory.update_state(node, Memory.STATE.DELETED)
                self._print_tick()
            elif Memory.STATE.STARTED == status:
                self._print_cross('The component must be stopped first')
                self._log.info('{} have to be stopped first'.format(node))
                break
            else:
                self._print_skip()
                self._log.info('skipped already deleted')
        else:
            self._print_loading_start('Delete network... ')
            docker_interface.delete_network(tpl.name)
            self._print_tick()
            shutil.rmtree(tpl.tmp_dir)

    def ls_components(self, app=None, filters={}):
        comps = Memory.get_comps(app, filters)

        def get_state(state):
            return colored(state, {
                Memory.STATE.CREATED.value: None,
                Memory.STATE.STARTED.value: 'green'}.get(state))

        def format_row(comp):
            return [comp['app_name'],
                    comp['name'],
                    comp['type'],
                    get_state(comp['state']),
                    '{}.{}'.format(comp['app_name'], comp['name'])]

        table = [format_row(c) for c in comps]
        table_str = tabulate(table, headers=['Application', 'Component',
                                             'Type', 'State', 'Full name'])
        Logger.println(table_str)

    def log(self, component, interface):
        # TODO: add logs also for Docker container
        try:
            split_list = component.split('.')
            app = '.'.join(split_list[:-1])
            name = split_list[-1]
        except ValueError:
            Logger.print_error('First argument must be a component full name (i.e my_app.my_component)')
            return
        
        self._log.debug('app: {}, name: {}, interface: {}'.format(app, name, interface))

        log_file_name = '{}/{}/*/{}/{}.log'.format(self._tmp_dir,
                                               app, name, interface)

        log_file = glob(log_file_name)

        if len(log_file) != 1:
            Logger.print_error('Component or interface log not found')
            return

        with open(log_file[0]) as f:
            for line in f.readlines():
                line = colored(line, 'green') if line.startswith('+ ') else line
                Logger.print_(line)

    def _print_outputs(self, tpl):
        if len(tpl.outputs) != 0:
            Logger.println('\nOUTPUTS:')
        for out in tpl.outputs:
            self._log.debug('value: {}'.format(out.value))
            value = out.value if isinstance(out.value, six.string_types) \
                else helper.get_attributes(out.value.args, tpl)
            Logger.println('  - ' + out.name + ":", value)

    def _update_state(self):
        errors = []

        def manage_error(comp, state):
            errors.append(comp['full_name'])
            Memory.update_state(comp, state)

        def manage_error_container(comp, state):
            manage_error(comp, state)
            path = os.path.join(self._tmp_dir, comp['app_name'], comp['name'])

            try:
                software = [(f.name, f.path) for f in scandir(path)
                            if f.is_dir()]
            except FileNotFoundError as e:
                software = []
            self._log.debug('path {} found {}'.format(path, software))

            for s, s_path in software:
                full_name = '{}.{}'.format(comp['app_name'], s)
                Memory.update_state('{}.{}'.format(comp['app_name'], s), state)
                # with open(os.path.join(s_path, 'state'), 'w') as f:
                #     f.write(state.value)
                try:
                    os.remove(os.path.join(s_path, 'state'))
                except FileNotFoundError:
                    pass
                errors.append(full_name)

        for c in Memory.get_comps(filters={'type': 'Software'}):
            state = glob('{}/{}/*/{}/state'.format(self._tmp_dir,
                                                   c['app_name'],
                                                   c['name']))
            self._log.debug('software update {}'.format(state))

            if len(state) == 1:
                with open(state[0], 'r') as f:
                    state = f.read().replace('\n', '')
                if state != c['state']:
                    manage_error(c, Memory.STATE(state))

        for c in Memory.get_comps(filters={'type': 'Container'}):
                status = docker_interface.inspect_container(c['full_name'])
                if status is not None:
                    if c['state'] == Memory.STATE.CREATED.value and \
                       status['State']['Running'] is not False:
                        manage_error_container(c, Memory.STATE.STARTED)
                    if c['state'] == Memory.STATE.STARTED.value and\
                       status['State']['Running'] is not True:
                        manage_error_container(c, Memory.STATE.CREATED)
                else:
                    manage_error_container(c, Memory.STATE.DELETED)

        for c in Memory.get_comps(filters={'type': 'Volume'}):
                status = docker_interface.inspect_volume(c['full_name'])
                if status is None:
                    manage_error(c, Memory.STATE.DELETED)

        return len(errors) == 0, errors

    def _components_exists(self, tpl, components):
        for c in components:
            if not any(c == n.name for n in tpl.nodes):
                return False
        return True

    def _extend_down(self, tpl, components):
        assert isinstance(components, list)

        def get_down_req(c):
            for r in c.relationships:
                if r.to.name not in components:
                    yield r.to
                for c in get_down_req(r.to):
                    yield c

        return self._extend(tpl, components, get_down_req)

    def _extend_up(self, tpl, components):
        assert isinstance(components, list)

        def get_up_req(c):
            for r in c.up_requirements:
                if r.origin.name not in components:
                    yield r.origin
                # python3 alternative
                # yield from get_up_req(r.origin)
                for c in get_up_req(r.origin):
                    yield c

        return self._extend(tpl, components, get_up_req)

    def _extend(self, tpl, components, extension_gen):
        assert isinstance(components, list)

        extend_comp = []

        for c in components:
            extend_comp.append(tpl[c])
            for rc in extension_gen(tpl[c]):
                extend_comp.append(rc)

        return extend_comp

    def _sort(self, tpl, components):
        assert isinstance(components, list)

        for n in tpl.nodes:
            n._mark = ''
        unmarked = set((c.name for c in components))
        deploy_order = []

        def visit(n):
            if n._mark == 'temp':
                self._log.debug('no dag')
                raise ValueError('the TOSCA file is not a DAG')
            elif n._mark == '':
                n._mark = 'temp'
                if n.name in unmarked:
                    unmarked.remove(n.name)
                for r in n.relationships:
                    if any(r.to == c for c in components):
                        visit(r.to)
                n._mark = 'perm'
                deploy_order.append(n)

        while len(unmarked) > 0:
            n = unmarked.pop()
            visit(tpl[n])

        return deploy_order

    def _print_tick(self):
        self._loading_thread.succeed(self._loading_thread.text + 'Done')

    def _print_skip(self):
        self._loading_thread.info(self._loading_thread.text + 'Skipped')

    def _print_cross(self, error):
        self._loading_thread.fail(self._loading_thread.text +
                                  'Error ({})'.format(error))

    def _print_loading_start(self, msg):
        self._loading_thread = Halo(text=msg, spinner='dots')
        self._loading_thread.start()



'''
Nodes module
'''
from .artifacts import Artifact, Dockerfile, DockerfileExecutable,\
                       DockerImage, DockerImageExecutable, File
from .relationships import AttachesTo, ConnectsTo, DependsOn, HostedOn


def _add_to_map(d, k, v):
    if d is None:
        d = {}
    d[k] = v
    return d


def _add_to_list(l_name, i):
    if l_name is None:
        l_name = []
    l_name.append(i)
    return l_name


def _str_obj(o):
    return ', '.join(["{}: {}".format(k, v) for k, v in vars(o).items()])


class Root(object):

    def __init__(self, name):
        self.name = name
        self.tpl = None

        self._ATTRIBUTE = {}

        # requirements
        self._depend = []
        self._connection = []
        self._volume = []
        self.artifacts = []

        # need by the topological sorting algorithm
        self._mark = ''

        # reverse requirements
        self.up_requirements = []

    @property
    def full_name(self):
        return '{}.{}'.format(self.tpl.name, self.name)

    @property
    def depend(self):
        return (i.format for i in self._depend)

    @property
    def connection(self):
        return (i.format for i in self._connection)

    @property
    def volume(self):
        return (i.format for i in self._volume)

    @property
    def relationships(self):
        return self._connection + self._volume + self._depend

    def add_depend(self, item):
        if not isinstance(item, DependsOn):
            item = DependsOn(self, item)
        self._depend.append(item)
        if not isinstance(item.to, str):
            item.to.up_requirements.append(item)

    def add_connection(self, item, alias=None):
        if not isinstance(item, ConnectsTo):
            item = ConnectsTo(self, item, alias)
        self._connection.append(item)
        if not isinstance(item.to, str):
            item.to.up_requirements.append(item)

    def add_volume(self, item, location=None):
        if not isinstance(item, AttachesTo):
            item = AttachesTo(self, item, location)
        self._volume.append(item)
        if not isinstance(item.to, str):
            item.to.up_requirements.append(item)

    def add_artifact(self, art):
        assert isinstance(art, Artifact)
        self.artifacts.append(art)

    def __str__(self):
        return self.name

    def __getitem__(self, item):
        return self._ATTRIBUTE.get(item, lambda: None)()

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)

    def get_str_obj(self):
        _str_obj(self)


class Container(Root):

    def __init__(self, name):
        super(self.__class__, self).__init__(name)
        # attributes
        self.id = None
        self.env = None
        self.cmd = None
        self.ports = None
        self.share_data = {}

        self._ATTRIBUTE = {
            'id': lambda: self.id,
            'ports': lambda: self.ports,
            'env_variable': lambda: self.env,
            'command': lambda: self.cmd,
            'share_data': lambda: self.share_data
        }
        self._overlay = []

        self.interfaces = {'create', 'start', 'stop', 'delete'}

    @property
    def image(self):
        return self.artifacts[0]

    @image.setter
    def image(self, img):
        assert isinstance(img, (DockerImage, Dockerfile))
        self.artifacts = [img]

    @property
    def executable(self):
        return isinstance(self.image, (DockerImageExecutable,
                                       DockerfileExecutable))

    @property
    def connection(self):
        return (i.format for i in self._connection + self._overlay)

    def add_overlay(self, item, alias=None):
        if not isinstance(item, ConnectsTo):
            item = ConnectsTo(self, item, alias)
        self._overlay.append(item)

    def add_env(self, name, value):
        self.env = _add_to_map(self.env, name, value)

    def add_port(self, name, value):
        self.ports = _add_to_map(self.ports, name, value)
    
    def add_share_data(self, name, value):
        self.share_data = _add_to_map(self.share_data, name, value)

    def get_str_obj(self):
        return '{}, {}'.format(
            super(self.__class__, self).__str__(), _str_obj(self)
        )

    def __str__(self):
        return '{} ({})'.format(self.name, 'container')


class Volume(Root):

    def __init__(self, name):
        super(self.__class__, self).__init__(name)
        # attributes
        self.id = None
        self.size = None
        self._ATTRIBUTE = {
            'id': lambda: self.id,
            'size': lambda: self.size
        }

        self.interfaces = {'create'
                           # , 'delete' TODO: add this interface
                           }

        self.driver_opt = None

    def get_all_opt(self):
        ris = self.driver_opt.copy() if self.driver_opt else {}
        if self.size:
            ris['size'] = self.size
        return ris

    def add_driver_opt(self, name, value):
        self.driver_opt = _add_to_map(self.driver_opt, name, value)

    def __str__(self):
        return '{} ({})'.format(self.name, 'volume')

    def get_str_obj(self):
        return '{}, {}'.format(super(self.__class__, self), _str_obj(self))


class Software(Root):

    def __init__(self, name):
        super(self.__class__, self).__init__(name)
        self.artifacts = []
        self.interfaces = {}

        # requirements
        self._host = None
        self.host_container = None
        # self.depend = None
        # self.connection = None

    @property
    def relationships(self):
        return super(self.__class__, self).relationships + \
               ([self._host] if self._host is not None else [])

    @property
    def host(self):
        return self._host

    @host.setter
    def host(self, item):
        if not isinstance(item, HostedOn):
            item = HostedOn(self, item)
        self._host = item
        if not isinstance(item.to, str):
            item.to.up_requirements.append(item)

    def add_artifact(self, file):
        assert isinstance(file, File)
        self.artifacts.append(file)

    def get_str_obj(self):
        return '{}, {}'.format(super(self.__class__, self), _str_obj(self))

    def __str__(self):
        return '{} ({})'.format(self.name, 'software')



import unittest

from tosker.orchestrator import Orchestrator

# from tosker import helper


class Test_Tosca_Parsing(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self._orchestrator = Orchestrator(
            # log_handler=helper.get_consol_handler()
        )

    def test_no_dag(self):
        self.assertFalse(
            self._orchestrator.orchestrate('tosker/tests/TOSCA/nodag.yaml',
                                           ['create'])
        )

    def test_wrong_type(self):
        self.assertFalse(
            self._orchestrator.orchestrate(
                'tosker/tests/TOSCA/wrong_components.yaml', [])
        )

    def test_wrong_component(self):
        self.assertFalse(
            self._orchestrator.orchestrate(
                'tosker/tests/TOSCA/wordpress.yaml', [], ['no_compent'], {})
        )


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import find_packages, setup

with open('DESCRIPTION.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

exec(open('tosker/__init__.py').read())

requirements = [  # 'Click>=6.0',
    'docker', 'tosca-parser', 'termcolor', 'six', 'enum34',
    'tabulate', 'tinydb', 'scandir', 'halo'
]

# put setup requirements (distutils extensions, etc.) here
setup_requirements = []

# put package test requirements here
test_requirements = []

# put package dev requirements here
dev_requirements = ['coverage', 'check-manifest', 'wheel', 'flake8', 'tox',
                    'Sphinx', 'isort']

setup(
    name='TosKer',
    version=__version__,
    description='A TOSCA engine working with Docker container',
    long_description=readme + '\n\n' + history,
    author="lucarin91",
    author_email='to@lucar.in',
    url='https://github.com/di-unipi-socc/TosKer',
    packages=find_packages(exclude=['tests']),
    entry_points={
        'console_scripts': [
            'tosker=tosker.ui:run'
        ]
    },
    include_package_data=True,
    data_files=[
        ('/usr/share/tosker', ['data/tosker-types.yaml']),
        ('/usr/share/tosker/examples',
            ['data/examples/thoughts-app/thoughts.csar',
             'data/examples/node-mongo-csar/node-mongo.csar'])

    ],
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='docker TOSCA deployment management development',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='tests',
    extras_require={'dev': dev_requirements},
    tests_require=test_requirements,
    setup_requires=setup_requirements
)
